#include <stdio.h>
#include <stdlib.h>
#include "ai.h"

/*Possible winning combinations
  WINS2 is defined for the first board.
  Add 9 to each value to get the second board and
  18 for the third.*/
int WINS2[8][3] = {		//add 9, then 18
    {0, 1, 2}, {3, 4, 5}, {6, 7, 8}, {0, 3, 6}, {1, 4, 7}, {2, 5, 8},
    {0, 4, 8}, {2, 4, 6}
};
/*Various diagonal and other 3d combinations*/
int WINS3[16][3] = {
    {0, 12, 24}, {1, 13, 25}, {2, 14, 26}, {6, 12, 18}, {7, 13, 19}, {8,
								      14,
								      20},
    {18, 13, 8}, {24, 13, 2}, {0, 13, 26}, {6, 13, 20}, {0, 10, 20}, {18,
								      10,
								      2},
    {3, 13, 23}, {21, 13, 5}, {6, 16, 26}, {24, 16, 8}
};
/*In addition, theres 0,9,18 1,10,19 etc...*/


/*allocates memory for a new board*/
BOARD *new_board()
{
    BOARD *result;
    int i;
    result = (BOARD *) malloc(SQUARES * sizeof(BOARD));
    for (i = 0; i < SQUARES; ++i)
	result[i] = ' ';
    return result;
}


/*Calculate the score given a board.
  ms stores the program's score
  os stores the player's score*/
void get_score(BOARD * b, int *ms, int *os)
{
    int i, j, z, me, opp;
    char d;
    (*ms) = (*os) = 0;
    /*Go through the various winning combinations */
    for (i = 0; i < 8; ++i)
	for (z = 0; z < 3; ++z) {
	    for (j = me = opp = 0; j < 3; ++j)
		if (b[WINS2[i][j] + z * 9] == x_)
		    opp++;
		else if (b[WINS2[i][j] + z * 9] == o_)
		    me++;
	    if (opp == 3)
		(*os)++;
	    else if (me == 3)
		(*ms)++;
	}
    for (i = 0; i < 16; ++i) {
	for (j = 0, me = opp = 0; j < 3; ++j)
	    if (b[WINS3[i][j]] == x_)
		opp++;
	    else if (b[WINS3[i][j]] == o_)
		me++;
	if (opp == 3)
	    (*os)++;
	else if (me == 3)
	    (*ms)++;
    }
    for (i = 0; i < 9; ++i) {
	for (j = me = opp = 0; j < 3; ++j) {
	    d = b[i + 9 * j];
	    if (d == x_)
		opp++;
	    else if (d == o_)
		me++;
	}
	if (opp == 3)
	    (*os)++;
	else if (me == 3)
	    (*ms)++;
    }
}



/*The heuristic function used to evaluate the board.
  For each winning 3 positions, we check to see if
  a)only one type of symbol (either x or o)
  b)if so, how many of that symbol are there.
  1 symbol :1 point
  2 symbols:2 points
  3 symbols:50 points (winning configuration)
  If the symbol is the opponent's then make the points negative (-1,-1,-50)*/
int evaluate(BOARD * b)
{
    int i, j, me, opp, z;
    int score = 0;
    for (i = 0; i < 8; ++i)
	for (z = 0; z < 3; ++z) {
	    for (j = 0, me = 0, opp = 0; j < 3; j++)
		if (b[WINS2[i][j] + z * 9] == x_)
		    opp++;
		else if (b[WINS2[i][j] + z * 9] == o_)
		    me++;
	    if (opp == 0)
		score += me;
	    else if (me == 0)
		score -= opp;
	}
    for (i = 0; i < 16; ++i) {
	for (j = 0, me = 0, opp = 0; j < 3; j++)
	    if (b[WINS3[i][j]] == x_)
		opp++;
	    else if (b[WINS3[i][j]] == o_)
		me++;
	if (opp == 0)
	    score += me;
	else if (me == 0)
	    score -= opp;
    }
    for (i = 0; i < 9; ++i) {
	for (j = 0, me = 0, opp = 0; j < 3; j++)
	    if (b[i + j * 9] == x_)
		opp++;
	    else if (b[i + j * 9] == o_)
		me++;
	if (opp == 0)
	    score += me;
	else if (me == 0)
	    score -= opp;
    }
    return score;
}

/*dump the board onto the screen*/
void print_board(BOARD * b)
{
    int i;
    for (i = 0; i < 3; ++i) {
	printf(" %c | %c | %c ", B(0 + i * 9), B(1 + i * 9), B(2 + i * 9));
	printf(" - + - + - \n");
	printf(" %c | %c | %c ", B(3 + i * 9), B(4 + i * 9), B(5 + i * 9));
	printf(" - + - + - \n");
	printf(" %c | %c | %c ", B(6 + i * 9), B(7 + i * 9), B(8 + i * 9));
	printf("\n");
    }
}


/*Alpha-Beta pruning*/
int prune(BOARD * b, int depth, int alpha, int beta, int turn)
{
    int v, i, c;
    v = evaluate(b);
#ifdef BOARD_DUMP
    printf("Depth=%i\tTurn=%i\tScore=%i\n", depth, turn, v);
    print_board(b);
#endif
    if (depth == 0) {
	return v;
    } else if (turn == MINE) {
	turn = OPP;
	for (c = 0, i = 0; i < SQUARES; i++)
	    if (b[i] == ' ') {
		b[i] = o_;
		c++;
		v = prune(b, depth - 1, alpha, beta, OPP);
		b[i] = ' ';
		alpha = max(alpha, v);
		if (alpha >= beta)
		    return beta;
	    }
	if (c == 0)
	    return evaluate(b);
	return alpha;
    } else			//opp move
    {
	turn = MINE;
	for (c = 0, i = 0; i < SQUARES; ++i)
	    if (b[i] == ' ') {
		c++;
		b[i] = x_;
		v = prune(b, depth - 1, alpha, beta, MINE);
		b[i] = ' ';
		beta = min(beta, v);
		if (alpha >= beta)
		    return alpha;
	    }
	if (c == 0)
	    return evaluate(b);
	return beta;
    }
}

/*Given a board, this function returns the move the program should make*/
int decide(BOARD * b)
{
    int scores[SQUARES];
    int i, mx, j;
    //turn=OPP;
    for (i = 0; i < SQUARES; ++i)
	scores[i] = -1000;
    for (i = 0; i < SQUARES; ++i)
	if (b[i] == ' ') {
	    b[i] = o_;
	    scores[i] = prune(b, DEPTH, -1000, 1000, OPP);
	    b[i] = ' ';
	}
    for (j = 0, i = 1, mx = scores[0]; i < SQUARES; ++i) {
	if (scores[i] > mx) {
	    j = i;
	    mx = scores[i];
	}
#ifdef BOARD_DUMP
	printf("%i:%i\n", i, scores[i]);
#endif
    }
    return j;
}
